from typing import Dict, List

import mlx.core as mx
import torch

from ..utils.weights import weights_to_mlx


def llama_to_mlxllm(weights_paths: List[str], verbose: bool = False) -> Dict[str, mx.array]:
    """Convert LLaMA-type weights to MLX format.

    Args:
        weights_paths (List[str]): list of paths to LLaMA-type weights
        verbose (bool, optional): whether to print information during conversion. Defaults to False.

    Returns:
        Dict[str, mx.array]: a dict of weights in MLX format
    """

    model_weights = {}
    weights = weights_to_mlx(weights_paths)
    if verbose:
        print("Converting LLaMA-type weights to mlx-llm format.")
    for k, w in weights.items():
        if k.startswith("model."):
            k = k.replace("model.", "")
        k_split = k.split(".")
        if "layers" in k_split:
            if "rotary_emb" in k_split:
                continue
            if "self_attn" in k_split:
                # ['layers', '0', 'attention', 'q_proj' | 'k_proj' | 'v_proj' | 'out_proj', weight]
                model_k = f"{k_split[0]}.{k_split[1]}.attention.{k_split[3]}.{k_split[4]}"
                model_weights[model_k] = w
            elif "mlp" in k_split:
                model_k = k
                model_weights[model_k] = w
            else:
                key_map = {
                    "input_layernorm": "attention_norm",
                    "post_attention_layernorm": "mlp_norm",
                    "pre_feedforward_layernorm": "pre_mlp_norm",
                    "post_feedforward_layernorm": "post_mlp_norm",
                }
                model_k = f"{k_split[0]}.{k_split[1]}.{key_map[k_split[2]]}.{k_split[3]}"
                model_weights[model_k] = w
        else:
            key_map = {"embed_tokens": "token_embed", "lm_head": "head", "norm": "norm"}
            model_k = f"{key_map[k_split[0]]}.{k_split[1]}"
            model_weights[model_k] = w
    return model_weights


def phi3_to_mlxllm(weights_paths: List[str], verbose: bool = False) -> Dict[str, mx.array]:
    """Convert original Phi3 weights to MLX format.

    Args:
        weights_paths (List[str]): list of paths to original Phi3 weights
        verbose (bool, optional): whether to print information during conversion. Defaults to False.

    Returns:
        Dict[str, mx.array]: a dict of weights in MLX format
    """
    model_weights = {}
    weights = weights_to_mlx(weights_paths)
    if verbose:
        print("Converting LLaMA 3 weights to mlx-llm format.")
    for k, w in weights.items():
        if k.startswith("model."):
            k = k.replace("model.", "")
        k_split = k.split(".")

        if "layers" in k_split:
            if "qkv_proj" in k_split:
                # ['layers', '0', 'attention', 'q_proj' | 'k_proj' | 'v_proj' | 'out_proj', weight]
                qkv = w.reshape(3, -1, w.shape[-1])
                q, k, v = qkv[0], qkv[1], qkv[2]
                q_key = f"{k_split[0]}.{k_split[1]}.attention.q_proj.{k_split[4]}"
                k_key = f"{k_split[0]}.{k_split[1]}.attention.k_proj.{k_split[4]}"
                v_key = f"{k_split[0]}.{k_split[1]}.attention.v_proj.{k_split[4]}"
                model_weights[q_key] = q
                model_weights[k_key] = k
                model_weights[v_key] = v
            elif "o_proj" in k_split:
                # - ['layers', '0', 'self_attn', 'o_proj', 'weight']
                model_k = f"{k_split[0]}.{k_split[1]}.attention.o_proj.{k_split[4]}"
                model_weights[model_k] = w
            elif "gate_up_proj" in k_split:
                gate_up_proj = w.reshape(2, -1, w.shape[-1])
                gate_proj, up_proj = gate_up_proj[0], gate_up_proj[1]
                gate_k = f"{k_split[0]}.{k_split[1]}.mlp.gate_proj.{k_split[4]}"
                up_k = f"{k_split[0]}.{k_split[1]}.mlp.up_proj.{k_split[4]}"
                model_weights[gate_k] = gate_proj
                model_weights[up_k] = up_proj
            elif "down_proj" in k_split:
                model_k = k
                model_weights[model_k] = w
            else:
                key_map = {"input_layernorm": "attention_norm", "post_attention_layernorm": "mlp_norm"}
                model_k = f"{k_split[0]}.{k_split[1]}.{key_map[k_split[2]]}.{k_split[3]}"
                model_weights[model_k] = w
        else:
            key_map = {"embed_tokens": "token_embed", "lm_head": "head", "norm": "norm"}
            model_k = f"{key_map[k_split[0]]}.{k_split[1]}"
            model_weights[model_k] = w
    return model_weights


def mistral_to_mlxllm(weights_paths: List[str], verbose: bool = False) -> Dict[str, mx.array]:
    """Convert Mistral weights to MLX format.

    Args:
        weights_paths (List[str]): list of paths to Mistral weights
        verbose (bool, optional): whether to print information during conversion. Defaults to False.

    Returns:
        Dict[str, mx.array]: a dict of weights in MLX format
    """
    model_weights = {}
    weights = weights_to_mlx(weights_paths)
    if verbose:
        print("Converting Mistral weights to mlx-llm format.")
    for k, w in weights.items():
        if k.startswith("model."):
            k = k.replace("model.", "")
        k_split = k.split(".")
        if "layers" in k_split:
            if "self_attn" in k_split:
                # ['layers', '0', 'self_attn', 'q_proj' | 'k_proj' | 'v_proj' | 'out_proj', weight]
                model_k = f"{k_split[0]}.{k_split[1]}.attention.{k_split[3]}.{k_split[4]}"
                model_weights[model_k] = w
            elif "mlp" in k_split:
                model_k = k
                model_weights[model_k] = w
            else:
                key_map = {"input_layernorm": "attention_norm", "post_attention_layernorm": "mlp_norm"}
                model_k = f"{k_split[0]}.{k_split[1]}.{key_map[k_split[2]]}.{k_split[3]}"
                model_weights[model_k] = w
        else:
            key_map = {"embed_tokens": "token_embed", "lm_head": "head", "norm": "norm"}
            model_k = f"{key_map[k_split[0]]}.{k_split[1]}"
            model_weights[model_k] = w
    return model_weights


def openelm_to_mlxllm(
    weights_paths: List[str],
    head_dim: int,
    n_kv_heads: List[int],
    tensor_type: torch.dtype = torch.float32,
    verbose: bool = False,
) -> Dict[str, mx.array]:
    """Convert OpenELM weights to MLX format.

    Args:
        weights_paths (List[str]): list of paths to OpenELM weights
        tensor_type (torch.dtype, optional): tensor type. Defaults to torch.float32.
        verbose (bool, optional): whether to print information during conversion. Defaults to False.

    Returns:
        Dict[str, mx.array]: a dict of weights in MLX format
    """
    model_weights = {}
    weights = weights_to_mlx(weights_paths, tensor_type=tensor_type)
    if verbose:
        print("Converting OpenELM weights to mlx-llm format.")
    for k, w in weights.items():
        if k.startswith("transformer."):
            k = k.replace("transformer.", "")
        k_split = k.split(".")
        if "qkv_proj" in k_split:
            idx = int(k_split[1])
            kv_dim = n_kv_heads[idx] * head_dim
            q_dim = w.shape[0] - 2 * kv_dim
            k_dim = kv_dim
            query, key, value = mx.split(w, [q_dim, q_dim + k_dim], axis=0)
            query_key = f"layers.{k_split[1]}.attention.q_proj.weight"
            key_key = f"layers.{k_split[1]}.attention.k_proj.weight"
            value_key = f"layers.{k_split[1]}.attention.v_proj.weight"
            model_weights[query_key] = query
            model_weights[key_key] = key
            model_weights[value_key] = value
        elif "out_proj" in k_split:
            model_key = f"layers.{k_split[1]}.attention.o_proj.weight"
            model_weights[model_key] = w
        elif "k_norm" in k_split or "q_norm" in k_split:
            model_key = f"layers.{k_split[1]}.attention.{k_split[3]}.weight"
            model_weights[model_key] = w
        elif "proj_1" in k_split:
            weight = w.reshape([2, -1, w.shape[-1]])
            gate_proj, up_proj = weight[0], weight[1]
            gate_key = f"layers.{k_split[1]}.mlp.gate_proj.weight"
            up_key = f"layers.{k_split[1]}.mlp.up_proj.weight"
            model_weights[gate_key] = gate_proj
            model_weights[up_key] = up_proj
        elif "proj_2" in k_split:
            model_key = f"layers.{k_split[1]}.mlp.down_proj.weight"
            model_weights[model_key] = w
        elif "attn_norm" in k_split:
            model_key = f"layers.{k_split[1]}.attention_norm.weight"
            model_weights[model_key] = w
        elif "ffn_norm" in k_split:
            model_key = f"layers.{k_split[1]}.mlp_norm.weight"
            model_weights[model_key] = w
        else:
            key_map = {"token_embeddings": "token_embed", "norm": "norm"}
            model_key = f"{key_map[k_split[0]]}.weight"
            model_weights[model_key] = w
    return model_weights



import time
from typing import Callable, Dict, Optional, Tuple

import mlx.core as mx
import mlx.nn as nn
import numpy as np
from mlx.optimizers import Optimizer
from tqdm import tqdm

from ..callbacks import ModelCheckpoint, StatsMonitor, TrainerStats
from ..loader import DataLoader
from ..model import get_weights, params_stats


def grad_checkpoint(model: nn.Module):
    """Update all instances of type(layer) to use gradient checkpointing.

    Args:
        model (nn.Module): model
    """
    fn = type(model).__call__

    def checkpointed_fn(model, *args, **kwargs):
        def inner_fn(params, *args, **kwargs):
            model.update(params)
            return fn(model, *args, **kwargs)

        return mx.checkpoint(inner_fn)(model.trainable_parameters(), *args, **kwargs)

    type(model).__call__ = checkpointed_fn


class Trainer:
    """Trainer class to train a model

    Args:
        model (nn.Module): the model to train
        optimizer (Optimizer): the optimizer to use
        train_loader (Dataset): the training loader
        val_loader (Optional[Dataset]): the validation loader. Defaults to None.
        max_epochs (int): the maximum number of epochs. Defaults to 10.
        val_every_n_epoch (int): validation frequency. Defaults to 1.
        log_every_n_steps (int): logging frequency. Defaults to 10.
        grad_ckpt (bool): whether to checkpoint gradients. If True, enables gradient checkpoint (some activations will not be saved but computed during backprop). Pro: spare some memory. Con: increased training time.
        model_checkpoint (Optional[ModelCheckpoint]): model checkpoint callback. Defaults to None.
        stats_monitor (Optional[StatsMonitor]): stats monitor callback. Defaults to None.
    """

    def __init__(
        self,
        model: nn.Module,
        optimizer: Optimizer,
        train_loader: DataLoader,
        val_loader: Optional[DataLoader] = None,
        max_epochs: int = 10,
        val_every_n_epoch: int = 1,
        stats_every_n_epoch: int = 1,
        log_every_n_steps: int = 10,
        grad_ckpt: bool = False,
        model_checkpoint: Optional[ModelCheckpoint] = None,
        stats_monitor: Optional[StatsMonitor] = None,
    ) -> None:
        self.model = model
        self.optimizer = optimizer
        self.train_loader = train_loader
        self.val_loader = val_loader
        self.max_epochs = max_epochs
        self.val_every_n_epoch = val_every_n_epoch
        self.stats_every_n_epoch = stats_every_n_epoch
        self.log_every_n_steps = log_every_n_steps
        self.grad_ckpt = grad_ckpt
        self.model_checkpoint = model_checkpoint
        self.stats_monitor = stats_monitor

        # define forward and backward step
        self.forward_pass = nn.value_and_grad(self.model, self._forward_pass)

        if self.grad_ckpt:
            grad_checkpoint(self.model)

        self._post_init()

    def _post_init(self) -> None:
        pass

    def _forward_pass(self, tokens: mx.array, target: mx.array, lengths: mx.array) -> Tuple[mx.array, int]:
        """Model forward pass

        Args:
            tokens (mx.array): input tokens
            target (mx.array): target tokens
            lengths (mx.array): lengths of the sequences

        Returns:
            Tuple[mx.array, int]: loss and number of tokens
        """
        # forward pass
        logits, _ = self.model(tokens)
        logits = logits.astype(mx.float32)

        # mask padding tokens
        length_mask = mx.arange(tokens.shape[1])[None, :] < lengths[:, None]

        # Cross entropy
        cross_entropy = nn.losses.cross_entropy(logits, tokens) * length_mask
        num_tokens = length_mask.sum()

        cross_entropy = cross_entropy.sum() / num_tokens
        return cross_entropy, num_tokens

    def train_epoch(self, trainer_stats: TrainerStats) -> TrainerStats:
        """Train the model for one epoch

        Args:
            trainer_stats (TrainerStats): stats of the epoch

        Returns:
            TrainerStats: stats of the epoch
        """
        self.model.train()
        for idx, (tokens, targets, lengths) in enumerate(self.train_loader):  # type: ignore
            iter_tic = time.perf_counter()
            # compute loss and gradients - forward pass
            (loss, num_tokens), gradients = self.forward_pass(tokens, targets, lengths)
            # update model - backward pass
            self.optimizer.update(self.model, gradients)
            mx.eval(self.model.parameters(), self.optimizer.state, loss)
            iter_toc = time.perf_counter()

            # update iter stats
            trainer_stats.train_iter_losses.append(loss.item())
            trainer_stats.train_iter_tokens.append(num_tokens.item())
            trainer_stats.train_iter_times.append(iter_toc - iter_tic)
            trainer_stats.train_iter_gpus.append(mx.metal.get_peak_memory() / 2**30)

            # log stats
            if idx % self.log_every_n_steps == 0:
                # log last iter stats
                log_loss = np.mean(trainer_stats.train_iter_losses[idx - self.log_every_n_steps :])
                log_tokens = np.mean(trainer_stats.train_iter_tokens[idx - self.log_every_n_steps :])
                log_time = np.sum(trainer_stats.train_iter_times[idx - self.log_every_n_steps :])
                np.mean(trainer_stats.train_iter_gpus[idx - self.log_every_n_steps :])
                print(
                    f"> Training Iter [{idx+1}/{len(self.train_loader)}] - loss: {log_loss:.4f} - tokens/s: {log_tokens/log_time:.2f}"
                )

        return trainer_stats

    def val_epoch(self, trainer_stats: Optional[TrainerStats] = None) -> Optional[TrainerStats]:  # type: ignore
        """Validation epoch

        Args:
            trainer_stats (Optional[TrainerStats]): stats of the epoch. Defaults to None.

        Returns:
            Optional[TrainerStats]: stats of the epoch
        """

        self.model.eval()
        val_loss, val_tokens = [], 0
        for _idx, (tokens, targets, lengths) in tqdm(
            enumerate(self.val_loader), total=len(self.val_loader), desc="> Running validation..."  # type: ignore
        ):
            # forward pass
            loss, num_tokens = self._forward_pass(tokens, targets, lengths)
            val_loss.append((loss * num_tokens).item())
            val_tokens += num_tokens.item()  # type: ignore

        val_loss = np.sum(val_loss) / val_tokens

        print(f"\n> Validation loss: {val_loss:.4f}")

        if trainer_stats is not None:
            trainer_stats.val_loss = val_loss  # type: ignore
            return trainer_stats

    # define a method that checks if it's time to run validation
    def _time_to_validate(self, epoch: int) -> bool:
        """Check if it's time to run validation

        Args:
            epoch (int): epoch number

        Returns:
            bool: True if it's time to run validation, False otherwise
        """
        return self.val_loader is not None and epoch % self.val_every_n_epoch == 0

    def _on_train_epoch_end(self, trainer_stats: TrainerStats) -> None:
        """Callbacks to run at the end of the training epoch

        Args:
            trainer_stats (TrainerStats): training stats
        """

        if self.model_checkpoint:
            self.model_checkpoint.step(trainer_stats=trainer_stats, weights=get_weights(self.model, trainable=True))

        if self.stats_monitor:
            self.stats_monitor.step(trainer_stats)

    def train(self) -> None:
        """Train the model"""
        for epoch in range(self.max_epochs):
            # init epoch stats
            trainer_stats = TrainerStats(epoch=epoch)

            print(f"\n******** Epoch {epoch}/{self.max_epochs} ********")

            epoch_tic = time.perf_counter()
            trainer_stats = self.train_epoch(trainer_stats=trainer_stats)
            time.perf_counter()

            trainer_stats.train_epoch_time = time.perf_counter() - epoch_tic
            print(f"> Epoch training time: {trainer_stats.train_epoch_time:.2f}s\n")

            # save epoch stats in output_dir
            if self._time_to_validate(epoch):
                trainer_stats = self.val_epoch(trainer_stats=trainer_stats)  # type: ignore

            self._on_train_epoch_end(trainer_stats)

    def eval(self) -> None:
        """Evaluate the model

        Raises:
            ValueError: Validation loader is required to evaluate the model
        """
        if self.val_loader is None:
            raise ValueError("Validation loader is required to evaluate the model")
        _ = self.val_epoch()



from ..utils.session import Session
from ._base import Prompt


class GemmaPrompt(Prompt):
    """Gemma prompt that follows this structure

    <bos><start_of_turn>user
    Write a hello world program<end_of_turn>
    <start_of_turn>model

    Args:
        system (str): system prompt
    """

    START = "<bos>"
    USER = "user"
    MODEL = "model"
    TURN_START = "<start_of_turn>"
    TEXT_END = "<end_of_turn>"

    def __init__(self, system: str) -> None:
        self.system = system

    def prepare(self, session: Session) -> str:
        """Prepare prompt for model input

        Args:
            session (Session): dialog session

        Returns:
            str: model input prompt
        """

        """
        <bos><start_of_turn>user
        {{ system }} + {{ question }} <end_of_turn>
        <start_of_turn>model
        """

        prompt = f"{self.START}{self.TURN_START}{self.USER}\n{self.system} "
        for i, qa in enumerate(session.history):
            # qa --> {"question": ..., "answer": ...}
            if i == 0:
                prompt += qa["question"] + " " + self.TEXT_END + "\n" + self.TURN_START + self.MODEL + "\n"  # type: ignore
            else:
                prompt += (
                    self.TURN_START  # type: ignore
                    + self.USER
                    + "\n"
                    + qa["question"]
                    + self.TEXT_END
                    + "\n"
                    + self.TURN_START
                    + self.MODEL
                    + "\n"
                )
            if qa["answer"] is None:
                break
            else:
                prompt += f"{qa['answer']} {self.TEXT_END}\n"

        return prompt



import json
import os
from typing import Any, Dict, List, Optional

import matplotlib
import matplotlib.pyplot as plt
import numpy as np

from .data import TrainerStats

matplotlib.use("Agg")


class StatsMonitor:
    """Monitor the training statistics and save them to disk.

    Args:
        output_dir (str): output directory
        plots_every_n_epoch (int): generate plots every n epoch
    """

    def __init__(self, output_dir: str, plots_every_n_epoch: int = 1):
        self.data_dir = os.path.join(output_dir, "data")
        self.plots_dir = os.path.join(output_dir, "plots")

        self.plots_every_n_epoch = plots_every_n_epoch
        self.stats_history: List[TrainerStats] = []

        # Create the output directories if they don't exist
        os.makedirs(self.data_dir, exist_ok=True)
        os.makedirs(self.plots_dir, exist_ok=True)

    def step(self, trainer_stats: TrainerStats) -> None:
        """Step the monitor.

        Args:
            trainer_stats (TrainerStats): trainer statistics
        """
        # Append the current epoch's statistics to the history
        self.stats_history.append(trainer_stats)

        # Save the current epoch's statistics
        self._save_epoch_stats(trainer_stats)

        # Generate plots every `plots_every_n_epoch`
        if trainer_stats.epoch % self.plots_every_n_epoch == 0:
            print(f"> Generating plots for epoch {trainer_stats.epoch} at {self.plots_dir}")
            self._make_plots()

    def _save_epoch_stats(self, trainer_stats: TrainerStats) -> None:
        """Save the epoch statistics to disk.

        Args:
            trainer_stats (TrainerStats): trainer statistics
        """
        # Save the stats in a JSON file, named by the epoch number
        epoch_filename = os.path.join(self.data_dir, f"epoch_{trainer_stats.epoch}.json")
        with open(epoch_filename, "w") as f:
            json.dump(trainer_stats.__dict__, f, indent=4)

    def _make_plots(self) -> None:
        """Generate the plots."""
        # Generate the plots
        train_iter_losses = []
        train_tokens_per_sec = []
        train_epoch_losses, val_losses = [], []
        epochs = []
        for stats in self.stats_history:
            train_iter_losses.extend(stats.train_iter_losses)
            train_tokens_per_sec.extend(stats.train_iter_tokens_per_sec)
            if stats.val_loss is not None:
                train_epoch_losses.append(np.mean(stats.train_iter_losses))
                val_losses.append(stats.val_loss)
                epochs.append(stats.epoch)

        # Plotting the training losses
        plt.figure()
        plt.plot(train_iter_losses, marker="o", color="orange", linestyle="-", markersize=3.0)
        plt.title("Training Losses")
        plt.xlabel("Iteration")
        plt.ylabel("Loss")
        plt.ylim(bottom=0, top=max(train_iter_losses) * 1.3)
        plt.savefig(os.path.join(self.plots_dir, "train_loss.png"))
        plt.close()

        # Plotting tokens per second
        plt.figure()
        plt.plot(train_tokens_per_sec, marker="o", color="orange")
        plt.title("Training Tokens/Second")
        plt.xlabel("Iteration")
        plt.ylabel("Tokens/Second")
        plt.ylim(bottom=0, top=max(train_tokens_per_sec) * 1.3)
        plt.savefig(os.path.join(self.plots_dir, "train_tokens_second.png"))
        plt.close()

        # Plotting validation loss if available
        if any(val_losses):
            # Plotting the comparison between training loss mean and validation loss
            plt.figure()
            plt.plot(
                epochs,
                train_epoch_losses,
                label="Training Loss Mean",
                marker="o",
                linestyle="-",
                markersize=3.0,
                color="orange",
            )
            plt.plot(
                epochs, val_losses, label="Validation Loss", marker="o", linestyle="-", markersize=3.0, color="blue"
            )
            plt.title("Training Loss (Mean) vs Validation Loss per Epoch")
            plt.xlabel("Epoch")
            plt.ylabel("Loss")
            plt.legend()
            plt.grid(True)
            plt.savefig(os.path.join(self.plots_dir, "train_vs_val_loss.png"))
            plt.close()



from typing import Any, Dict, Iterable

import mlx.core as mx
import mlx.nn as nn
from mlx.utils import tree_unflatten

from .layers import LoRALinear


def _to_lora(layer: nn.Module, lora_config: Dict[str, Any]) -> nn.Module:
    """Converts a layer to a LoRA layer.

    Args:
        layer (nn.Module): a layer to convert.
        lora_config (Dict[str, Any]): LoRA configuration.

    Raises:
        ValueError: if the layer is not supported.

    Returns:
        nn.Module: a LoRA layer.
    """
    if isinstance(layer, (nn.Linear, nn.QuantizedLinear)):
        return LoRALinear(layer, **lora_config)  # type: ignore
    else:
        raise ValueError(f"Layer {layer} is not supported for LoRA conversion.")


def convert_layers(
    model: nn.Module,
    lora_layers_keys: Iterable[str] = ("attention.q_proj", "attention.k_proj", "attention.v_proj"),
    num_lora_layers: int = 4,
    lora_config: Dict[str, Any] = {"r": 8, "dropout": 0.2, "alpha": 32, "bias": False},  # noqa: B006
) -> nn.Module:
    """Converts the last `num_lora_layers` layers of the model to LoRA layers.

    Args:
        model (nn.Module): a model to convert.
        lora_layers_keys (Iterable[str], optional): keys of the layers to convert. Defaults to ("attention.q_proj", "attention.k_proj", "attention.v_proj").
        num_lora_layers (int, optional): number of layers to convert. Defaults to 16.
        lora_config (Dict[str, Any], optional): LoRA configuration. Defaults to DEFAULT_LORA_CONFIG.

    Returns:
        nn.Module: a model with LoRA layers.
    """

    num_model_layers = len(model.layers)
    if num_lora_layers > num_model_layers or num_lora_layers < 0:
        raise ValueError(
            f"num_lora_layers ({num_lora_layers}) must be less than the number of layers in the model ({num_model_layers}) and >0."
        )

    start_from = num_model_layers - num_lora_layers
    applied = 0

    if isinstance(lora_layers_keys, str):
        lora_layers_keys = [lora_layers_keys]

    for layer in model.layers[start_from:]:
        lora_layers = [(k, _to_lora(m, lora_config)) for k, m in layer.named_modules() if k in lora_layers_keys]
        if lora_layers:
            applied += 1
            layer.update_modules(tree_unflatten(lora_layers))

    if applied == 0:
        raise ValueError(f"No LoRA layers found in the last {num_lora_layers} layers.")
    else:
        print(f"Converted last {applied} layers with keys ({lora_layers_keys}) for LoRA.")

    return model



from queue import Queue
from threading import Thread
from typing import Any, Callable

import mlx.core as mx
import numpy as np
from transformers import PreTrainedTokenizer

from ..dataset import Dataset


# TODO: must be improved: add sampler support and other features
class DataLoader:
    """Custom DataLoader (similar to PyTorch but easier to read).

    Args:
        dataset (Dataset): the dataset to be loaded.
        tokenizer (PreTrainedTokenizer): the tokenizer to use.
        batch_size (int): the batch size. Defaults to 1.
        shuffle (bool): whether to shuffle the dataset. Defaults to False.
        collate_fn (Callable): the function to use to collate the data. Defaults to _default_collate_fn.
        drop_last (bool): whether to drop the last batch if it's smaller than the batch size. Defaults to False.
    """

    def __init__(
        self,
        dataset: Dataset,
        tokenizer: PreTrainedTokenizer,
        batch_size: int = 1,
        max_length: int = 2048,
        shuffle: bool = False,
        drop_last: bool = False,
    ):
        self.dataset = dataset
        self.tokenizer = tokenizer
        self.batch_size = batch_size
        self.max_length = max_length
        self.shuffle = shuffle
        self.indices = list(range(len(self.dataset)))

        # if drop_last is true, we need to make sure that the last batch is not smaller than the batch size
        if drop_last:
            self.indices = self.indices[: len(self.indices) - len(self.indices) % self.batch_size]

    def __iter__(self) -> Any:
        if self.shuffle:
            np.random.shuffle(self.indices)
        for idx in range(0, len(self.indices), self.batch_size):
            batch_indices = self.indices[idx : idx + self.batch_size]
            batch = self.tokenizer(
                [self.dataset[i] for i in batch_indices],
                max_length=self.max_length + 1,
                truncation=True,
                padding="max_length",
            )
            batch = mx.array(batch["input_ids"])
            lengths = mx.array([len(self.dataset[i]) for i in batch_indices])
            # yielding :
            # * batch[:, :-1] -> the input sequence (all but the last token)
            # * batch[:, 1:] -> the target sequence (all but the first token)
            # * lengths -> the length of each sequence
            yield batch[:, :-1], batch[:, 1:], mx.array(lengths)

    def __len__(self) -> int:
        return len(self.indices) // self.batch_size



import os
from dataclasses import dataclass
from typing import Dict, List, Optional, Tuple

import mlx.core as mx

from mlx_llm.utils.weights import save_weights

from .data import TrainerStats


class ModelCheckpoint:
    """Model checkpoint class to save checkpoints and monitoring them. Filename for each checkpoint will be filled with trainer_stats from Trainer.

    Args:
        output_dir (str): output dir where checkpoints folder will contain mlx files.
        monitor (str, optional): metric to monitor. Defaults to "loss".
        mode (str, optional): mode to check metric to monitor. Defaults to "min".
        save_top_k (int, optional): how many top mlx files to keep during training. Defaults to 5.
        patience (int, optional): how many epochs to wait before ending training if metric to monitor does not improve. Defaults to 10.
        weights_ext (str, optional): extension for weights file. Defaults to ".safetensors".
    """

    def __init__(
        self,
        output_dir: str,
        monitor: str = "val_loss",
        mode: str = "min",
        save_top_k: int = 5,
        patience: int = 10,
        weights_ext: str = ".safetensors",
    ) -> None:
        assert mode in ["max", "min"], f"Mode {mode} not supported. Choose between max or min."
        assert (
            monitor in TrainerStats.__dataclass_fields__
        ), f"Monitor {monitor} not in TrainerStats attributes. Valid attributes are {TrainerStats.__annotations__}"

        self.output_dir = os.path.join(output_dir, "adapters")
        os.makedirs(self.output_dir, exist_ok=True)
        self.monitor = monitor
        self.history: List[Tuple[float, int]] = []  # each el is a Tuple of (val, epoch)
        self.save_top_k = save_top_k
        self.mode = mode
        self.reverse = True if self.mode == "max" else False
        self.patience = patience
        self.weights_ext = weights_ext

        # support fields
        self.patience_count = 0
        self.to_remove = None  # an epoch will be saved here

    def _update_history_sequence(self, value: float, epoch: int, remove_last: bool) -> None:
        """Update history sequence

        Args:
            value (float): value to add
            epoch (int): current epoch
            remove_last (bool): remove last from history
        """

        if remove_last:
            self.to_remove = self.history[-1][1]  # type: ignore
            self.history = self.history[:-1]

        self.history.append((value, epoch))
        self.history = sorted(self.history, reverse=self.reverse, key=lambda x: x[0])
        self.patience_count = 0

    @property
    def best_value(self) -> Optional[float]:
        """Return best value of history sequence

        Returns:
            Optional[float]: best value
        """
        if len(self.history) == 0:
            return None
        return sorted(self.history, reverse=self.reverse, key=lambda x: x[0])[0][0]

    def _update_history(self, trainer_stats: TrainerStats) -> None:
        """Update history

        Args:
            trainer_stats (TrainerStats): TrainerStats instance
        """

        value = trainer_stats.__dict__[self.monitor]
        epoch = trainer_stats.epoch

        if len(self.history) < self.save_top_k:
            self._update_history_sequence(value=value, epoch=epoch, remove_last=False)
            print(
                f"> Epoch {epoch} with best model with {self.monitor}={value:.4f}. Best model is at {self.monitor}={self.best_value:.4f}"
            )
        else:
            # checking for history full
            to_update = True
            if self.mode == "max" and value < min(self.history, key=lambda x: x[0])[0]:
                to_update = False
            elif self.mode == "min" and value > max(self.history, key=lambda x: x[0])[0]:
                to_update = False

            if not to_update:
                print(
                    f"> Epoch {epoch} was not between best models with {self.monitor}={value:.4f}. Current best model interval is {self.monitor}={self.best_value:.4f}"
                )
                self.patience_count += 1
            else:
                self._update_history_sequence(value=value, epoch=epoch, remove_last=True)

    @property
    def patience_over(self) -> bool:
        """Check if patience is over

        Returns:
            bool: True if patience over, False otherwise
        """
        return self.patience_count >= self.patience

    def _create_filename(self, trainer_stats: TrainerStats) -> str:
        """Create mlx filename

        Args:
            trainer_stats (TrainerStats): TrainerStats instance

        Returns:
            str: model filename
        """

        filename = f"epoch={trainer_stats.epoch}-"
        value = trainer_stats.__dict__[self.monitor]
        filename += f"{self.monitor}={value:.4f}"
        return f"{filename}{self.weights_ext}"

    def save(self, weights: Dict, filename: str) -> None:
        """Save mlx file and removes old worst one if needed

        Args:
            weights (Dict): model weights
            filename (str): filename to save
        """

        if len(self.history) == self.save_top_k:
            for f in os.listdir(self.output_dir):
                if self.to_remove is not None:
                    if f.startswith(f"epoch={self.to_remove}"):  # type: ignore
                        os.remove(os.path.join(self.output_dir, f))
        try:
            save_weights(weights, os.path.join(self.output_dir, filename))
            print(f"> Saved model mlx file ({filename}) in checkpoints folder.")
        except Exception as e:
            print(f"[ERROR] Error while saving model. Error {e}.")

    def step(self, trainer_stats: TrainerStats, weights: Dict[str, mx.array]) -> None:
        """Update model checkpoint data and save mlx file if needed

        Args:
            epoch (int): current epoch
            trainer_stats (TrainerStats): trainer stats
            weights (Dict): model weights
        """

        trainer_stats.__dict__[self.monitor]

        self._update_history(trainer_stats=trainer_stats)
        # it means one of best models just added
        if self.patience_count == 0:
            filename = self._create_filename(trainer_stats=trainer_stats)
            self.save(weights=weights, filename=filename)



from ..utils.session import Session
from ._base import Prompt


class MistralPrompt(Prompt):
    """Mistral Instruct prompt that follows this structure

    [INST] {{ system_prompt }} + {{ user_prompt }} [/INST] {{ model_prompt }}</s>[INST] {{ user_prompt }} [/INST]

    Args:
        system (str): system prompt

    """

    INST_START = "[INST]"
    INST_END = "[/INST]"
    TEXT_START = "<s>"
    TEXT_END = "</s>"

    def __init__(self, system: str) -> None:
        self.system = system

    def prepare(self, session: Session) -> str:
        """Prepare prompt for model input

        Args:
            session (Session): dialog session

        Returns:
            str: model input prompt
        """

        """
        [INST] {{ system_prompt }} {{ user_prompt }} [/INST] {{ model_prompt }}</s>[INST] {{ user_prompt }} [/INST]
        """
        """
        "<s>[INST] What is your favourite condiment? [/INST]"
        "Well, I'm quite partial to a good squeeze of fresh lemon juice. It adds just the right amount of zesty flavour to whatever I'm cooking up in the kitchen!</s> "
        "[INST] Do you have mayonnaise recipes? [/INST]"
        """

        prompt = f"{self.TEXT_START} {self.INST_START} {self.system} "
        for i, qa in enumerate(session.history):
            # qa --> {"question": ..., "answer": ...}
            if i == 0:
                prompt += f"{qa['question']} {self.INST_END}\n"
            else:
                prompt += f"{self.INST_START} {qa['question']} {self.INST_END}\n"
            if qa["answer"] is None:
                break
            else:
                prompt += f"{qa['answer']} {self.TEXT_END}\n"

        return prompt


class StarlingPrompt(Prompt):
    """Starling Chat prompt that follows this structure

    GPT4 Correct User: {prompt}<|end_of_turn|>GPT4 Correct Assistant: {response}<|end_of_turn|>GPT4 Correct User: {follow_up_question}<|end_of_turn|>GPT4 Correct Assistant:

    Args:
        system (str): system prompt
    """

    USER = "GPT4 Correct User:"
    ASSISTANT = "GPT4 Correct Assistant:"
    TEXT_END = "<|end_of_turn|>"

    def __init__(self, system: str) -> None:
        self.system = system

    def prepare(self, session: Session) -> str:
        """Prepare prompt for model input

        Args:
            session (Session): dialog session

        Returns:
            str: model input prompt
        """

        """
        GPT4 Correct User: {prompt}<|end_of_turn|>GPT4 Correct Assistant: {response}<|end_of_turn|>GPT4 Correct User: {follow_up_question}<|end_of_turn|>GPT4 Correct Assistant:
        """

        prompt = f"{self.USER} {self.system}"
        for i, qa in enumerate(session.history):
            if i == 0:
                prompt += f"{qa['question']}{self.TEXT_END}"
            else:
                prompt += f"{self.USER} {qa['question']}{self.TEXT_END}"
            if qa["answer"] is None:
                prompt += f"{self.ASSISTANT}"
            else:
                prompt += f"{self.ASSISTANT} {qa['answer']} {self.TEXT_END}"

        return prompt



from typing import Optional

from huggingface_hub import hf_hub_download, snapshot_download


def download_from_hf(repo_id: str, revision: Optional[str] = None, filename: Optional[str] = None) -> str:
    """Download entire repo from HuggingFace Hub.

    Args:
        repo_id (str): HuggingFace Hub repo id
        revision (Optional[str], optional): An optional Git revision id which can be a branch name, a tag, or a commit hash.
        filename (Optional[str], optional): An optional filename to download from the repo.

    Returns:
        str: path to downloaded weights
    """
    try:
        if filename is None:
            model_path = snapshot_download(
                repo_id=repo_id,
                revision=revision,
                allow_patterns=[
                    "*.json",
                    "*.safetensors",
                    "tokenizer.model",
                    "*.tiktoken",
                ],
                resume_download=True,
            )
        else:
            model_path = hf_hub_download(repo_id=repo_id, repo_type="model", filename=filename)
    except Exception as e:
        print(f"[ERROR] Downloading repo from HuggingFace Hub failed: {e}.")
        raise e

    return model_path



from typing import Dict, List, Optional, Tuple, Union

import mlx.core as mx
import mlx.nn as nn


class RMSNorm(nn.Module):
    """Custom RMSNorm module

    Args:
        dims (int): model dimension
        eps (float, optional): epsilon. Defaults to 1e-5.
    """

    def __init__(self, dims: int, eps: float = 1e-5) -> None:
        super().__init__()
        self.weight = mx.ones((dims,))
        self.eps = eps

    def __call__(self, x: mx.array) -> mx.array:
        """Forward pass

        Args:
            x (mx.array): input

        Returns:
            mx.array: output
        """
        return mx.fast.rms_norm(x, 1.0 + self.weight, self.eps)


class Attention(nn.Module):
    """Attention module

    Args:
        dim (int): model dimension
        n_heads (int): number of attention heads
        n_kv_heads (int): number of key-value heads
        head_dim (Optional[int]): head dimension. If None, it is set to dim // n_heads. Defaults to None.
        rope_traditional (bool, optional): whether to use traditional RoPE. Defaults to False.
        rope_theta (float, optional): RoPE theta. Defaults to 1000.
        rope_scaling (Optional[Dict[str, Union[float, str]]], optional): RoPE scaling. Defaults to None.
        norm_qk_proj (bool, optional): whether to normalize the queries and keys projection (for OpenELM). Defaults to False.
        attention_norm_eps (float, optional): attention normalization epsilon (for OpenELM). Defaults to 1e-6.
    """

    def __init__(
        self,
        dim: int,
        n_heads: int,
        n_kv_heads: int,
        head_dim: Optional[int] = None,
        rope_traditional: bool = False,
        rope_theta: float = 1000,
        rope_scaling: Optional[Dict[str, Union[float, str]]] = None,
        norm_qk_proj: bool = False,
        attention_norm_eps: float = 1e-6,
    ) -> None:
        super().__init__()

        self.n_heads = n_heads
        self.n_kv_heads = n_kv_heads
        self.norm_qk_proj = norm_qk_proj

        head_dim = dim // n_heads if head_dim is None else head_dim

        self.scale = head_dim**-0.5

        self.q_proj = nn.Linear(dim, n_heads * head_dim, bias=False)
        self.k_proj = nn.Linear(dim, n_kv_heads * head_dim, bias=False)
        self.v_proj = nn.Linear(dim, n_kv_heads * head_dim, bias=False)
        self.o_proj = nn.Linear(n_heads * head_dim, dim, bias=False)

        if self.norm_qk_proj:
            self.q_norm = nn.RMSNorm(dims=head_dim, eps=attention_norm_eps)
            self.k_norm = nn.RMSNorm(dims=head_dim, eps=attention_norm_eps)

        rope_scale = 1 / rope_scaling["factor"] if rope_scaling is not None and rope_scaling["type"] == "linear" else 1  # type: ignore

        self.rope = nn.RoPE(
            dims=head_dim,
            traditional=rope_traditional,
            base=rope_theta,
            scale=rope_scale,
        )

    def __call__(
        self,
        x: mx.array,
        mask: Optional[mx.array] = None,
        kv_cache: Optional[Tuple[mx.array, mx.array]] = None,
    ) -> Tuple[mx.array, Tuple[mx.array, mx.array]]:
        """Forward pass

        Args:
            x (mx.array): input tokens
            mask (Optional[mx.array], optional): attention mask. Defaults to None.
            kv_cache (Optional[Tuple[mx.array, mx.array]], optional): key-value cache. Defaults to None.

        Returns:
            Tuple[mx.array, Tuple[mx.array, mx.array]]: output and kv-cache
        """
        B, L, D = x.shape

        queries, keys, values = self.q_proj(x), self.k_proj(x), self.v_proj(x)

        # Prepare the queries, keys and values for the attention computation
        queries = queries.reshape(B, L, self.n_heads, -1).transpose(0, 2, 1, 3)
        keys = keys.reshape(B, L, self.n_kv_heads, -1).transpose(0, 2, 1, 3)
        values = values.reshape(B, L, self.n_kv_heads, -1).transpose(0, 2, 1, 3)

        if self.norm_qk_proj:
            queries = self.q_norm(queries)
            keys = self.k_norm(keys)

        if kv_cache is not None:
            key_cache, value_cache = kv_cache
            queries = self.rope(queries, offset=key_cache.shape[2])
            keys = self.rope(keys, offset=key_cache.shape[2])
            keys = mx.concatenate([key_cache, keys], axis=2)
            values = mx.concatenate([value_cache, values], axis=2)
        else:
            queries = self.rope(queries)
            keys = self.rope(keys)
        output = mx.fast.scaled_dot_product_attention(queries, keys, values, scale=self.scale, mask=mask)  # type: ignore
        output = output.transpose(0, 2, 1, 3).reshape(B, L, -1)
        return self.o_proj(output), (keys, values)


class MLP(nn.Module):
    """MLP module

    Args:
        dim (int): model dimension
        hidden_dim (int): hidden dimension
        gemma (bool, optional): use Gemma activation. Defaults to False.
    """

    def __init__(self, dim: int, hidden_dim: int, gemma: bool = False):
        super().__init__()

        self.gate_proj = nn.Linear(dim, hidden_dim, bias=False)
        self.down_proj = nn.Linear(hidden_dim, dim, bias=False)
        self.up_proj = nn.Linear(dim, hidden_dim, bias=False)
        self.gemma = gemma

    def __call__(self, x: mx.array) -> mx.array:
        """Forward pass

        Args:
            x (mx.array): input

        Returns:
            mx.array: output
        """
        if self.gemma:
            return self.down_proj(nn.gelu(self.gate_proj(x)) * self.up_proj(x))
        else:
            return self.down_proj(nn.silu(self.gate_proj(x)) * self.up_proj(x))


class TransformerBlock(nn.Module):
    """Transformer block

    Args:
        dim (int): model dimension
        n_heads (int): number of attention heads
        n_kv_heads (int): number of key-value heads
        hidden_dim (int): hidden dimension
        norm_eps (float): normalization epsilon
        head_dim (Optional[int], optional): head dimension. Defaults to None.
        rope_traditional (bool, optional): whether to use traditional RoPE. Defaults to False.
        rope_theta (float, optional): RoPE theta. Defaults to 1000.
        rope_scaling (Optional[Dict[str, Union[float, str]]], optional): RoPE scaling. Defaults to None.
        norm_qk_proj (bool, optional): whether to normalize the queries and keys projection (for OpenELM). Defaults to False.
        attention_norm_eps (float, optional): attention normalization epsilon (for OpenELM). Defaults to 1e-6.
        gemma (bool, optional): whether using Gemma layers. Defaults to False.
        gemma2 (bool, optional): whether using Gemma2 layers. Defaults to False.
    """

    def __init__(
        self,
        dim: int,
        n_heads: int,
        n_kv_heads: int,
        hidden_dim: int,
        norm_eps: float,
        head_dim: Optional[int] = None,
        rope_traditional: bool = False,
        rope_theta: float = 1000,
        rope_scaling: Optional[Dict[str, Union[float, str]]] = None,
        norm_qk_proj: bool = False,
        attention_norm_eps: float = 1e-6,
        gemma: bool = False,
        gemma2: bool = False,
    ) -> None:
        super().__init__()
        self.n_heads = n_heads
        self.dim = dim
        self.attention = Attention(
            dim=dim,
            n_heads=n_heads,
            n_kv_heads=n_kv_heads,
            head_dim=head_dim,
            rope_traditional=rope_traditional,
            rope_theta=rope_theta,
            rope_scaling=rope_scaling,
            norm_qk_proj=norm_qk_proj,
            attention_norm_eps=attention_norm_eps,
        )
        self.mlp = MLP(dim=dim, hidden_dim=hidden_dim, gemma=gemma)
        self.gemma2 = gemma2
        if not gemma:
            self.attention_norm = nn.RMSNorm(dim, eps=norm_eps)
            self.mlp_norm = nn.RMSNorm(dim, eps=norm_eps)
        else:
            self.attention_norm = RMSNorm(dim, eps=norm_eps)  # type: ignore
            self.mlp_norm = RMSNorm(dim, eps=norm_eps)  # type: ignore
            if gemma2:
                self.pre_mlp_norm = RMSNorm(dim, eps=norm_eps)  # type: ignore
                self.post_mlp_norm = RMSNorm(dim, eps=norm_eps)  # type: ignore

    def __call__(
        self,
        x: mx.array,
        mask: Optional[mx.array] = None,
        kv_cache: Optional[Tuple[mx.array, mx.array]] = None,
    ) -> Tuple[mx.array, Tuple[mx.array, mx.array]]:
        """Forward pass

        Args:
            x (mx.array): input tokens
            mask (Optional[mx.array], optional): attention mask. Defaults to None.
            kv_cache (Optional[Tuple[mx.array, mx.array]], optional): key-value cache. Defaults to None.

        Returns:
            Tuple[mx.array, Tuple[mx.array, mx.array]]: output and key-value cache
        """
        r, kv_cache = self.attention(x=self.attention_norm(x), mask=mask, kv_cache=kv_cache)
        if self.gemma2:
            h = x + self.mlp_norm(r)
        else:
            h = x + r
        if self.gemma2:
            r = self.mlp(self.pre_mlp_norm(h))
            r = self.post_mlp_norm(r)
        else:
            r = self.mlp(self.mlp_norm(h))
        out = h + r
        return out, kv_cache


class Transformer(nn.Module):
    """Transformer model

    Args:
        dim (int): model dimension
        hidden_dim (int): hidden dimension
        vocab_size (int): vocabulary size
        n_layers (int): number of transformer blocks
        n_heads (int): number of attention heads
        n_kv_heads (Optional[int]): number of key-value heads. If None, it is set to num_heads. Defaults to None.
        head_dim (Optional[int], optional): head dimension. Defaults to None.
        norm_eps (float): normalization epsilon. Defaults to 1e-5.
        rope_traditional (bool, optional): whether to use traditional RoPE. Defaults to False.
        rope_theta (float, optional): RoPE theta. Defaults to 1000.
        rope_scaling (Optional[Dict[str, Union[float, str]]], optional): RoPE scaling. Defaults to None.
        norm_qk_proj (bool, optional): whether to normalize the queries and keys projection (for OpenELM). Defaults to False.
        attention_norm_eps (float, optional): attention normalization epsilon (for OpenELM). Defaults to 1e-6.
        gemma (bool, optional): whether to use Gemma Transformer. Defaults to False.
        gemma2 (bool, optional): whether to use Gemma2 Transformer. Defaults to False.
    """

    def __init__(
        self,
        dim: int,
        hidden_dim: Union[int, List[int]],
        vocab_size: int,
        n_layers: int,
        n_heads: Union[int, List[int]],
        n_kv_heads: Optional[Union[int, List[int]]] = None,
        head_dim: Optional[int] = None,
        norm_eps: float = 1e-5,
        rope_traditional: bool = False,
        rope_theta: float = 1000,
        rope_scaling: Optional[Dict[str, Union[float, str]]] = None,
        norm_qk_proj: bool = False,
        attention_norm_eps: float = 1e-6,
        gemma: bool = False,
        gemma2: bool = False,
        embed_as_head: bool = False,
    ):
        super().__init__()

        assert vocab_size > 0

        self.dim = dim
        self.hidden_dim = hidden_dim
        self.vocab_size = vocab_size
        self.n_layers = n_layers
        self.gemma = gemma
        self.gemma2 = gemma2
        if gemma2 and not gemma:
            raise ValueError("Gemma2 can only be used with Gemma")
        self.embed_as_head = True if self.gemma is True else embed_as_head

        if n_kv_heads is None:
            n_kv_heads = n_heads

        self.token_embed = nn.Embedding(vocab_size, self.dim)

        if isinstance(hidden_dim, int):
            hidden_dim = [hidden_dim] * n_layers
        if isinstance(n_heads, int):
            n_heads = [n_heads] * n_layers
        if isinstance(n_kv_heads, int):
            n_kv_heads = [n_kv_heads] * n_layers

        self.layers = [
            TransformerBlock(
                dim=dim,
                n_heads=n_heads[i],
                n_kv_heads=n_kv_heads[i],
                hidden_dim=hidden_dim[i],
                norm_eps=norm_eps,
                head_dim=head_dim,
                rope_traditional=rope_traditional,
                rope_theta=rope_theta,
                rope_scaling=rope_scaling,
                norm_qk_proj=norm_qk_proj,
                attention_norm_eps=attention_norm_eps,
                gemma=self.gemma,
                gemma2=self.gemma2,
            )
            for i in range(n_layers)
        ]
        self.norm = nn.RMSNorm(dim, eps=norm_eps) if not self.gemma else RMSNorm(dim, eps=norm_eps)
        self.head = nn.Linear(dim, vocab_size, bias=False) if not self.embed_as_head else None  # type: ignore

    def embed(
        self, x: mx.array, kv_cache: Optional[List[Tuple[mx.array, mx.array]]] = None, norm: bool = False
    ) -> Tuple[mx.array, Optional[List[Tuple[mx.array, mx.array]]]]:
        """Compute embedding for the input tokens.

        Args:
            x (mx.array): input tokens
            kv_cache (Optional[List[Tuple[mx.array, mx.array]]]): key-value cache
            norm (bool, optional): whether to normalize the output. Defaults to False.

        Returns:
            Tuple[mx.array, Optional[List[Tuple[mx.array, mx.array]]]]: output and key-value cache
        """
        h = self.token_embed(x)

        if self.gemma:
            h = h * (self.dim**0.5)

        mask = None
        if h.shape[1] > 1:
            mask = nn.MultiHeadAttention.create_additive_causal_mask(h.shape[1])
            mask = mask.astype(h.dtype)

        if kv_cache is None:
            kv_cache = [None] * len(self.layers)  # type: ignore

        for e, layer in enumerate(self.layers):
            h, kv_cache[e] = layer(x=h, mask=mask, kv_cache=kv_cache[e])

        return self.norm(h) if norm else h, kv_cache

    def __call__(
        self, x: mx.array, kv_cache: Optional[List[Tuple[mx.array, mx.array]]] = None
    ) -> Tuple[mx.array, List[Tuple[mx.array, mx.array]]]:
        """Forward pass

        Args:
            x (mx.array): input tokens
            kv_cache (Optional[List[mx.array]], optional): key-value cache. Defaults to None.

        Returns:
            Tuple[mx.array, List[mx.array]]: output and key-value cache
        """
        x, kv_cache = self.embed(x, kv_cache=kv_cache, norm=True)
        if self.embed_as_head:
            out = self.token_embed.as_linear(x)
        else:
            out = self.head(x)  # type: ignore
        return out, kv_cache  # type: ignore

    def generate(self, x: mx.array, temp: Optional[float] = 0.0):
        """Generate tokens from a given input

        Args:
            x (mx.array): input tokens
            temp (Optional[float], optional): model temperature. Defaults to 0.0.
        """

        def sample(logits):
            if temp == 0:
                return mx.argmax(logits, axis=-1)
            else:
                return mx.random.categorical(logits * (1 / temp))

        logits, kv_cache = self(x[None])
        y = sample(logits[:, -1, :])
        yield y

        while True:
            logits, kv_cache = self(y[:, None], kv_cache=kv_cache)
            y = sample(logits.squeeze(1))
            yield y



STOP = 0
WAIT = 1
GO = 2


def answer_status(text: str, end_str: str) -> int:
    """Check if the model is done generating the end_str

    Args:
        text (str): generated text
        end_str (str): end string

    Returns:
        int: STOP, WAIT, or GO
    """
    if text[-len(end_str) :] == end_str:
        return STOP
    for i in range(len(end_str) - 1, 0, -1):
        if text[-i:] == end_str[:i]:
            return WAIT
    return GO



from dataclasses import dataclass, field
from typing import List, Optional


@dataclass
class TrainerStats:  # noqa: D101
    epoch: int
    train_iter_losses: List[float] = field(default_factory=list)
    train_iter_tokens: List[int] = field(default_factory=list)
    train_iter_times: List[float] = field(default_factory=list)
    train_iter_gpus: List[float] = field(default_factory=list)
    train_epoch_time: float = 0.0

    val_loss: Optional[float] = None

    @property
    def train_iter_tokens_per_sec(self) -> List[float]:
        """Compute the tokens per second for each training iteration

        Returns:
            List[float]: tokens per second for each training iteration
        """
        return [tokens / time for tokens, time in zip(self.train_iter_tokens, self.train_iter_times)]



import glob
import os
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import mlx.core as mx
import mlx.nn as nn
from transformers import AutoTokenizer

from ..utils.hf import download_from_hf
from ._registry import MODEL_ENTRYPOINTS
from ._utils import apply_weights, load_weights, quantize

__all__ = ["list_models", "create_model"]


def list_models() -> List[str]:
    """List all available LLM models."""
    all_models = list(MODEL_ENTRYPOINTS.keys())
    return all_models


def create_model(
    model_name: str,
    weights: Union[List[str], str, bool] = True,
    strict: bool = False,
    converter: Optional[Callable] = None,
    verbose: bool = False,
    model_config: Optional[Dict[str, Any]] = None,
) -> nn.Module:
    """Create a LLM model.

    Example:

    ```
    >>> from mlx_llm.model import create_model

    >>> # Create a LLaMA 3 8B model with pretrained weights from HF.
    >>> model = create_model('llama_3_8b_instruct')

    >>> # Create a LLaMA 3 8B model with pretrained weights from HF from another repository.
    >>> model = create_model(
            model_name="llama_3_8b_instruct", # it's the base model
            weights="hf://gradientai/Llama-3-8B-Instruct-262k", # new weights from HuggingFace
            model_config={ "rope_theta": 207112184.0 }
    >>> )

    >>> # Create a LLaMA 3 8B model model with custom local weights.
    >>> model = create_model('llama_2_8b_instruct', weights="path/to/model.safetensors")
    ```

    Args:
        model_name (str): model name
        weights (Union[str, List[str], bool]): if True, load pretrained weights from HF. If str or List[str], load weights from the given paths. Defaults to True.
        strict (bool, optional): whether to strictly enforce that the keys in weights match the keys of the model. Defaults to False.
        converter (Optional[Callable], optional): a function to convert the weights to the model format. Defaults to None.
        verbose (bool, optional): whether to print the model summary. Defaults to False.
        model_config (Dict[str, Any], optional): model configuration. Defaults to {}.

    Returns:
        nn.Module: a LLM model

    Raises:
        ValueError: Unknown model name
    """

    if model_name not in MODEL_ENTRYPOINTS:
        raise ValueError(f"Unknown model name: {model_name}.")

    if model_config is None:
        model_config = {}
    model, config = MODEL_ENTRYPOINTS[model_name](**model_config)

    if config.quantize is not None:
        model = quantize(model, group_size=config.quantize.group_size, bits=config.quantize.bits)

    if isinstance(weights, str) and weights.endswith(".safetensors"):
        model = load_weights(model=model, weights=weights, strict=strict, verbose=verbose)
    elif isinstance(weights, str) and weights.startswith("hf://"):
        model_path = download_from_hf(
            repo_id=weights.replace("hf://", ""),
        )
        weights = glob.glob(os.path.join(model_path, "*.safetensors"))  # type: ignore
        if len(weights) == 0:
            raise ValueError(f"No safetensors weights found in {model_path}.")
        converter = config.converter if converter is None else converter
        weights = converter(weights)
        model = apply_weights(model, weights)  # type: ignore

    elif isinstance(weights, bool) and weights is True:
        model_path = download_from_hf(
            repo_id=config.hf.repo_id, revision=config.hf.revision, filename=config.hf.filename
        )
        if os.path.isfile(model_path):
            weights = model_path
        else:
            weights = glob.glob(os.path.join(model_path, "*.safetensors"))  # type: ignore
        if config.converter is not None:
            weights = config.converter(weights)
            model = apply_weights(model, weights)  # type: ignore

    return model


def create_tokenizer(model_name: Optional[str] = None, path: Optional[str] = None) -> AutoTokenizer:
    """Create a tokenizer for a LLM model.

    Args:
        model_name (Optional[str], optional): model name. Defaults to None.
        path (Optional[str], optional): path to local tokenizer. Defaults to None.

    Raises:
        ValueError: Unknown model name

    Returns:
        AutoTokenizer: tokenizer
    """
    # if both model_name and path are None, raise an error
    if model_name is None and path is None:
        raise ValueError("Either model_name or path must be provided.")

    # if path is given, load the tokenizer from the path
    if path is not None and os.path.isdir(path):
        return AutoTokenizer.from_pretrained(path, legacy=True)

    if model_name is not None:
        if model_name.startswith("hf://"):
            repo_id = model_name.replace("hf://", "")
        elif model_name in MODEL_ENTRYPOINTS:
            _, config = MODEL_ENTRYPOINTS[model_name]()
            repo_id = config.tokenizer.repo_id if config.tokenizer else config.hf.repo_id
        else:
            raise ValueError(f"Unknown model name: {model_name}.")

        return AutoTokenizer.from_pretrained(repo_id, legacy=True)

    raise ValueError("Invalid arguments: model_name or path must be provided.")



from ..utils.session import Session
from ._base import Prompt


class OpenELMPrompt(Prompt):
    """OpenELM instruct prompt that follows this structure

    <|system|>
    You are a helpful AI assistant. </s>
    <|user|>
    How to explain Internet for a medieval knight? </s>
    <|assistant|>

    Args:
        system (str): system prompt

    """

    SYSTEM_START = "<|system|>"
    USER = "<|user|>"
    ASSISTANT = "<|assistant|>"
    TEXT_END = "</s>"

    def __init__(self, system: str) -> None:
        self.system = system

    def prepare(self, session: Session) -> str:
        """Prepare prompt for model input

        Args:
            session (Session): dialog session

        Returns:
            str: model input prompt
        """

        """
        <|system|>
        You are a helpful AI assistant.</s>
        <|user|>
        How to explain Internet for a medieval knight?</s>
        <|assistant|>
        """

        prompt = f"{self.SYSTEM_START}\n{self.system}{self.TEXT_END}"
        for qa in session.history:
            # qa --> {"question": ..., "answer": ...}
            prompt += f"\n{self.USER}\n{qa['question']}{self.TEXT_END}"
            if qa["answer"] is None:
                break
            else:
                prompt += f"\n{self.ASSISTANT}\n{qa['answer']}{self.TEXT_END}"

        prompt += f"\n{self.ASSISTANT}\n"

        return prompt



from ..utils.session import Session
from ._base import Prompt


class DeepSeekLLaMA3Prompt(Prompt):
    """DeepSeek R1 LLaMA 3.1 prompt that follows this structure
    <｜begin▁of▁sentence｜>{system_prompt}<｜User｜>{prompt}<｜Assistant｜>{assistant}<｜end▁of▁sentence｜><｜User｜>{prompt}<｜Assistant｜>{assistant}<｜end▁of▁sentence｜>

    Args:
        system (str): system prompt

    """

    PROMPT_START = "<｜begin▁of▁sentence｜>"
    TEXT_END = "<｜end▁of▁sentence｜>"
    USER = "<｜User｜>"
    ASSISTANT = "<｜Assistant｜>"

    def __init__(self, system: str) -> None:
        self.system = system

    def prepare(self, session: Session) -> str:
        """Prepare prompt for model input

        Args:
            session (Session): dialog session

        Returns:
            str: model input prompt
        """

        """
        <｜begin▁of▁sentence｜>{system_prompt}<｜User｜>{prompt}<｜Assistant｜>{assistant}<｜end▁of▁sentence｜><｜User｜>{prompt}<｜Assistant｜>{assistant}<｜end▁of▁sentence｜>
        """

        prompt = f"{self.PROMPT_START}{self.system}"
        for qa in session.history:
            # qa --> {"question": ..., "answer": ...}
            prompt += f"{self.USER}{qa['question']}{self.ASSISTANT}"
            if qa["answer"] is None:
                break
            else:
                prompt += f"{qa['question']}{self.TEXT_END}"

        return prompt



import json
from typing import Dict, List, Optional

from transformers import PreTrainedTokenizer


class Dataset:
    """Base class for datasets.

    Args:
        data_path (str): path to the data file.
        data_key (str): key of the data in the file. Defaults to "text".
        n_samples (int): number of samples to load for debugging. Defaults to 100.
    """

    def __init__(self, data_path: str, data_key: str = "text", n_samples: Optional[int] = None) -> None:
        with open(data_path, "r") as r:
            self.data = [json.loads(l) for l in r]
            if n_samples:
                self.data = self.data[:n_samples]
        self.data_key = data_key

    def __getitem__(self, idx: int) -> str:
        return self.data[idx][self.data_key]

    def __len__(self) -> int:
        return len(self.data)


class ChatDataset(Dataset):
    """Dataset for chat data.

    Args:
        data_path (str): path to the data file.
        tokenizer (PreTrainedTokenizer): tokenizer to use.
        n_samples (int): number of samples to load for debugging. Defaults to 100.
    """

    def __init__(self, data_path: str, tokenizer: PreTrainedTokenizer, n_samples: Optional[int] = None) -> None:
        super().__init__(data_path, n_samples=n_samples)
        self.tokenizer = tokenizer

    def __getitem__(self, idx: int) -> str:
        messages = self.data[idx]["messages"]
        text = self.tokenizer.apply_chat_template(messages, tokenize=False, add_generation_prompt=True)
        return text



import os
import time
from typing import Dict, List, Optional, Tuple, Union

import mlx.core as mx
import mlx.nn as nn
import numpy as np
from mlx.utils import tree_flatten, tree_unflatten
from transformers import AutoTokenizer


def apply_weights(
    model: nn.Module,
    weights: Dict[str, mx.array],  # type: ignore
) -> nn.Module:
    """Apply weights to a model.

    Args:
        model (nn.Module): a model
        weights (Dict[str, mx.array]): weights dict

    Returns:
        nn.Module: a model with weights applied
    """
    model.update(tree_unflatten(list(weights.items())))
    return model


def load_weights(
    model: nn.Module, weights: Union[str, List[str]], strict: bool = True, verbose: bool = False
) -> nn.Module:
    """Load weights from a given path.

    Args:
        model (nn.Module): a LLM model
        weights (Union[str, List[str]]): path to weights file or list of paths to weights files
        strict (bool, optional): whether to strictly enforce that the keys in weights match the keys of the model. Defaults to True.
        verbose (bool, optional): whether to print information during loading. Defaults to False.

    Returns:
        nn.Module: an nn.Module with loaded weights
    """

    if verbose:
        print(f"\n> Loading weights from {weights}")

    # loading weights
    if isinstance(weights, str):
        weights = [weights]
    for weight in weights:
        assert os.path.exists(weight), f"Weights path {weight} does not exist."

    pretrained_weights = {}
    for weight in weights:
        pretrained_weights.update(dict(list(mx.load(weight).items())))

    # create a torch-like state dict { layer_name: weights }
    model_weights = dict(tree_flatten(model.parameters()))

    if strict:
        if len(model_weights) != len(pretrained_weights):
            raise ValueError(f"Expected {len(model_weights)} keys, got {len(pretrained_weights)}")
        if set(model_weights.keys()) != set(pretrained_weights.keys()):
            diff = set(model_weights.keys()) ^ set(pretrained_weights.keys())
            raise ValueError(f"Found model keys not in pretrained weights: {diff}")
        if set(pretrained_weights.keys()) != set(model_weights.keys()):
            diff = set(pretrained_weights.keys()) ^ set(model_weights.keys())
            raise ValueError(f"Found pretrained keys not in model weights: {diff}")

    for k, w in model_weights.items():
        if k not in pretrained_weights:
            if strict:
                raise KeyError(f"Missing key {k} in weights file")
            elif verbose:
                print(f"> [WARNING] Missing key {k} in weights file")
            continue
        else:
            pretrained_w = pretrained_weights[k]
            if verbose:
                print(f"> [SUCCESS] Found key {k} in weights file")
            # checking if pretrained_w has the same shape as w
            if pretrained_w.shape != w.shape:
                if strict:
                    raise ValueError(f"Expected shape {w.shape} for key {k}, got {pretrained_w.shape}")
                elif verbose:
                    print(f"> [WARNING] Expected shape {w.shape} for key {k}, got {pretrained_w.shape}")
                    pretrained_w = w
            model_weights[k] = pretrained_w

    model = apply_weights(model, model_weights)
    return model


def get_weights(model: nn.Module, trainable: bool = False) -> dict:
    """Return the model weights dict. If trainable is True, return only trainable parameters.

    Args:
        model (nn.Module): a model
        trainable (bool, optional): whether to return trainable parameters. Defaults to False.

    Returns:
        dict: model weights dict
    """
    if trainable:
        state_dict = dict(tree_flatten(model.trainable_parameters()))
    else:
        state_dict = dict(tree_flatten(model.parameters()))
    return state_dict


def quantize(model: nn.Module, group_size: int, bits: int) -> nn.Module:
    """
    Applies quantization to the model weights.

    Args:
        model (nn.Module): model to be quantized.
        group_size (int): group size for quantization.
        bits (int): bits per weight for quantization.

    Returns:
        Tuple: model
    """

    nn.quantize(model, group_size, bits)  # type: ignore
    return model


def generate(
    model: nn.Module,
    tokenizer: AutoTokenizer,
    prompt: str,
    max_tokens: int = 100,
    temperature: float = 0.1,
    stats: bool = True,
) -> None:
    """Generate text from a given prompt.

    Args:
        model (nn.Module): model
        tokenizer (AutoTokenizer): tokenizer
        prompt (str): prompt
        max_tokens (int, optional): max number of tokens to generate. Defaults to 100.
        temperature (float, optional): temperature. Defaults to 0.1.
    """
    # generate answer
    x = mx.array([tokenizer.bos_token_id] + tokenizer.encode(prompt))

    skip = 0
    tokens = []

    print(prompt, end="", flush=True)

    for _i, token in enumerate(model.generate(x, temperature)):
        if _i == 0:
            tick = time.time()
        tokens.append(token.item())  # actually compute the token
        if len(tokens) >= max_tokens:
            break
        token_list = list(tokens)
        if token_list[-1] == tokenizer.eos_token_id:
            break
        answer = tokenizer.decode(token_list)
        print(answer[skip:], end="", flush=True)
        skip = len(answer)
    tock = time.time()
    tokens_per_second = len(tokens) / (tock - tick)
    if stats:
        print(f"\n\n[STATS] Tokens/sec: {tokens_per_second:.3f}")


def make_divisible(
    v: Union[float, int],
    divisor: Optional[int] = 8,
    min_value: Optional[Union[float, int]] = None,
) -> Union[float, int]:
    """
    This function is taken from the original tf repo.
    It ensures that all layers have a channel number that is divisible by the divisor
    It can be seen at:
    https://github.com/tensorflow/models/blob/2cfc99eff5e5eb729c6793d2f3d03aa1c9be2b15/research/slim/nets/mobilenet/mobilenet.py#L62
    Args:
        v: input value
        divisor: default to 8
        min_value: minimum divisor value
    Returns:
        new_v: new divisible value
    """
    if min_value is None:
        min_value = divisor
    new_v = max(min_value, int(v + divisor / 2) // divisor * divisor)  # type: ignore
    # Make sure that round down does not go down by more than 10%.
    if new_v < 0.9 * v:  # type: ignore
        new_v += divisor  # type: ignore
    return new_v  # type: ignore



import os
from pathlib import Path
from typing import Dict, Optional

import yaml


def load_config(config_path: str) -> Dict:
    """Load a single yml file.

    Args:
        config_path (str): path to yml file

    Raises:
        FileExistsError: if config file does not exist

    Returns:
        Dict: yml dict
    """
    if os.path.exists(path=config_path) is False:
        raise FileExistsError(f"Config file {config_path} does not exist.")

    with open(config_path) as stream:
        try:
            params = yaml.safe_load(stream)
        except yaml.YAMLError as exc:
            print(exc)
            quit()
    return params



from ._base import Prompt
from .deepseek import DeepSeekLLaMA3Prompt
from .gemma import GemmaPrompt
from .hermes import HermesPrompt
from .llama import LLaMA2Prompt, LLaMA3Prompt, TinyLLaMAPrompt
from .mistral import MistralPrompt, StarlingPrompt
from .openelm import OpenELMPrompt
from .phi import Phi3Prompt

PROMPT_ENTRYPOINTS = {
    "llama_2_7b_chat_hf": LLaMA2Prompt,
    "llama_2_7b_hf": LLaMA2Prompt,
    "llama_3_8b": LLaMA3Prompt,
    "llama_3_8b_instruct": LLaMA3Prompt,
    "llama_3_2_1b_instruct": LLaMA3Prompt,
    "llama_3_2_3b_instruct": LLaMA3Prompt,
    "hermes_2_pro_llama_3_8b": HermesPrompt,
    "phi_3_mini_4k_instruct": Phi3Prompt,
    "phi_3_mini_128k_instruct": Phi3Prompt,
    "phi_3.5_mini_instruct": Phi3Prompt,
    "mistral_7b_instruct_v0.2": MistralPrompt,
    "starling_lm_7b_beta": StarlingPrompt,
    "openhermes_2.5_mistral_7b": HermesPrompt,
    "tiny_llama_1.1B_chat_v1.0": TinyLLaMAPrompt,
    "gemma_1.1_2b_it": GemmaPrompt,
    "gemma_1.1_7b_it": GemmaPrompt,
    "gemma_2_2b_it": GemmaPrompt,
    "gemma_2_9b_it": GemmaPrompt,
    "openelm_3B_instruct": OpenELMPrompt,
    "openelm_1.1B_instruct": OpenELMPrompt,
    "openelm_450M_instruct": OpenELMPrompt,
    "openelm_270M_instruct": OpenELMPrompt,
    "smollm2_1.7B_instruct": HermesPrompt,
    "smollm2_360M_instruct": HermesPrompt,
    "smollm2_135M_instruct": HermesPrompt,
    "deepseek_r1_distill_llama_8b": DeepSeekLLaMA3Prompt,
}


def list_prompts() -> None:
    """List all available prompts."""
    print("Available prompts based on model families:")
    for model_family in list(PROMPT_ENTRYPOINTS.keys()):
        print(f"\t- {model_family}")


def create_prompt(model_name: str, system: str = "") -> Prompt:
    """Create prompt based on model family

    Args:
        model_name (str): model name
        system (str): system prompt

    Returns:
        Prompt: prompt class
    """

    assert model_name in PROMPT_ENTRYPOINTS.keys(), f"Model {model_name} not found in available prompts."
    return PROMPT_ENTRYPOINTS[model_name](system=system)  # type: ignore



from ..utils.session import Session
from ._base import Prompt


class HermesPrompt(Prompt):
    """Hermes instruct prompt that follows this structure

    <|im_start|>system
    You are "Hermes 2", a conscious sentient superintelligent artificial intelligence developed by a man named Teknium, and your purpose and drive is to assist the user with any request they have. You experience emotions and have deep, profound thoughts and qualia.<|im_end|>
    <|im_start|>user
    Hello, who are you?<|im_end|>
    <|im_start|>assistant
    Hi there! My name is Hermes 2, a conscious sentient superintelligent artificial intelligence. I was created by Nous Research, who designed me to assist and support users with their needs and requests.<|im_end|>

    Args:
        system (str): system prompt

    """

    TEXT_START = "<|im_start|>"
    USER = "user"
    SYSTEM = "system"
    ASSISTANT = "assistant"
    TEXT_END = "<|im_end|>"

    def __init__(self, system: str) -> None:
        self.system = system

    def prepare(self, session: Session) -> str:
        """Prepare prompt for model input

        Args:
            session (Session): dialog session

        Returns:
            str: model input prompt
        """

        """
        <|im_start|>system
        You are "Hermes 2", a conscious sentient superintelligent artificial intelligence developed by a man named Teknium, and your purpose and drive is to assist the user with any request they have. You experience emotions and have deep, profound thoughts and qualia.<|im_end|>
        <|im_start|>user
        Hello, who are you?<|im_end|>
        <|im_start|>assistant
        Hi there! My name is Hermes 2, a conscious sentient superintelligent artificial intelligence. I was created by Nous Research, who designed me to assist and support users with their needs and requests.<|im_end|>
        """

        prompt = f"{self.TEXT_START}{self.SYSTEM}\n{self.system}{self.TEXT_END}\n"
        for qa in session.history:
            # qa --> {"question": ..., "answer": ...}
            prompt += f"{self.TEXT_START}{self.USER}\n{qa['question']}{self.TEXT_END}\n"
            if qa["answer"] is None:
                prompt += f"{self.TEXT_START}{self.ASSISTANT}\n"
            else:
                prompt += f"{self.TEXT_START}{self.ASSISTANT}\n{qa['answer']}{self.TEXT_END}\n"
        if len(session) == 0:
            prompt += f"{self.TEXT_START}{self.ASSISTANT}\n"
        return prompt



import math
from typing import Optional

import mlx.core as mx
import mlx.nn as nn


class LoRALinear(nn.Module):
    """LoRA Linear layer.

    The output is calculated as:
    .. math::
        \text{Output} = (W_0 \times x) + \alpha \\cdot (A \times B \\times x)

    Where:
    - `x` is the input to the layer.
    - `W_0` is the original weight matrix from the pre-trained model.
    - `A` and `B` are the low-rank matrices used for the adaptation.
    - `alpha` is the scaling factor applied to the low-rank update.


    Args:
        layer (nn.Linear, optional): pre-trained linear layer. Defaults to None.
        input_dims (int): input dimensions. Defaults to None.
        output_dims (int): output dimensions. Defaults to None.
        r (int, optional): rank of the low-rank matrices used in LoRA. Defaults to 8.
        drop (float, optional): dropout rate. Defaults to 0.0.
        alpha (float, optional): scale factor. Defaults to 20.0.
        bias (bool, optional): use bias. Defaults to False.
    """

    def __init__(
        self,
        layer: Optional[nn.Linear] = None,
        input_dims: Optional[int] = None,
        output_dims: Optional[int] = None,
        r: int = 8,
        dropout: float = 0.2,
        alpha: float = 32.0,
        bias: bool = False,
    ):
        assert (layer is not None) or (
            input_dims is not None and output_dims is not None
        ), "Either layer or input_dims and output_dims must be provided."

        super().__init__()
        if layer is not None:
            output_dims, input_dims = layer.weight.shape
            self.layer = layer
        else:
            if input_dims is None or output_dims is None:
                raise ValueError("input_dims and output_dims must be provided.")
            self.layer = nn.Linear(input_dims, output_dims, bias=bias)  # type: ignore

        self.dropout = nn.Dropout(p=dropout)
        # scale for low-rank update
        self.alpha = alpha
        # A init values --> random uniform around 1/sqrt(input_dims) in (-val, val) -> shape is (input_dims, r)
        val = 1 / math.sqrt(input_dims)  # type: ignore
        self.a = mx.random.uniform(low=-val, high=val, shape=(input_dims, r))
        self.b = mx.zeros(shape=(r, output_dims))

    def to_linear(self):  # noqa: D102
        linear = self.layer
        bias = "bias" in linear
        weight = linear.weight
        is_quantized = isinstance(linear, nn.QuantizedLinear)

        # Use the same type as the linear weight if not quantized
        dtype = weight.dtype

        if is_quantized:
            dtype = mx.float16
            weight = mx.dequantize(
                weight,
                linear.scales,
                linear.biases,
                linear.group_size,
                linear.bits,
            )
        output_dims, input_dims = weight.shape
        fused_linear = nn.Linear(input_dims, output_dims, bias=bias)

        lora_b = (self.alpha * self.b.T).astype(dtype)
        lora_a = self.a.T.astype(dtype)
        fused_linear.weight = weight + lora_b @ lora_a
        if bias:
            fused_linear.bias = linear.bias

        if is_quantized:
            fused_linear = nn.QuantizedLinear.from_linear(
                fused_linear,
                linear.group_size,
                linear.bits,
            )

        return fused_linear

    def __call__(self, x: mx.array) -> mx.array:
        """Forward pass.

        Args:
            x (mx.array): input tensor

        Returns:
            mx.array: output tensor
        """
        # linear layer
        y = self.layer(x)
        # low-rank update
        z = (self.dropout(x) @ self.a) @ self.b
        out = y + (self.alpha * z).astype(x.dtype)
        return out



from ..utils.session import Session
from .hermes import HermesPrompt
from .llama import TinyLLaMAPrompt


class Phi3Prompt(TinyLLaMAPrompt):
    """Phi3 Instruct prompt that follows this structure

    <|system|>
    You are a helpful AI assistant.<|end|>
    <|user|>
    How to explain Internet for a medieval knight?<|end|>
    <|assistant|>

    Args:
        system (str): system prompt

    """

    SYSTEM_START = "<|system|>"
    USER = "<|user|>"
    ASSISTANT = "<|assistant|>"
    TEXT_END = "<|end|>"

    def __init__(self, system: str) -> None:
        super().__init__(system)


class Phi4Prompt(HermesPrompt):
    """Phi4 prompt that follows this structure

    <|im_start|>system<|im_sep|>
    You are a medieval knight and must provide explanations to modern people.<|im_end|>
    <|im_start|>user<|im_sep|>
    How should I explain the Internet?<|im_end|>
    <|im_start|>assistant<|im_sep|>
    """

    TEXT_START = "<|im_start|>"
    TEXT_END = "<|im_end|>"
    TEXT_SEP = "<|im_sep|>"
    USER = "user"
    SYSTEM = "system"
    ASSISTANT = "assistant"

    def __init__(self, system: str) -> None:
        self.system = system

    def prepare(self, session: Session) -> str:
        """Prepare prompt for model input

        Args:
            session (Session): dialog session

        Returns:
            str: model input prompt
        """

        prompt = f"{self.TEXT_START}{self.SYSTEM}{self.TEXT_SEP}\n{self.system}{self.TEXT_END}\n"
        for qa in session.history:
            prompt += f"{self.TEXT_START}{self.USER}{self.TEXT_SEP}\n{qa['question']}{self.TEXT_END}\n"
            if qa["answer"] is None:
                prompt += f"{self.TEXT_START}{self.ASSISTANT}{self.TEXT_SEP}\n"
            else:
                prompt += f"{self.TEXT_START}{self.ASSISTANT}{self.TEXT_SEP}\n{qa['answer']}{self.TEXT_END}\n"
        if len(session) == 0:
            prompt += f"{self.TEXT_START}{self.ASSISTANT}{self.TEXT_SEP}\n"
        return prompt



from abc import ABC, abstractmethod

from ..utils.session import Session


class Prompt(ABC):
    TEXT_END: str = ""

    def __init__(self, system: str) -> None:
        self.system = system
        pass

    @abstractmethod
    def prepare(self, session: Session) -> str:
        pass
